/* USER CODE BEGIN Header */
/**
  ******************************************************************************
  * @file           : main.c
  * @brief          : Main program body
  ******************************************************************************
  ** This notice applies to any and all portions of this file
  * that are not between comment pairs USER CODE BEGIN and
  * USER CODE END. Other portions of this file, whether 
  * inserted by the user or by software development tools
  * are owned by their respective copyright owners.
  *
  * COPYRIGHT(c) 2020 STMicroelectronics
  *
  * Redistribution and use in source and binary forms, with or without modification,
  * are permitted provided that the following conditions are met:
  *   1. Redistributions of source code must retain the above copyright notice,
  *      this list of conditions and the following disclaimer.
  *   2. Redistributions in binary form must reproduce the above copyright notice,
  *      this list of conditions and the following disclaimer in the documentation
  *      and/or other materials provided with the distribution.
  *   3. Neither the name of STMicroelectronics nor the names of its contributors
  *      may be used to endorse or promote products derived from this software
  *      without specific prior written permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
  * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
  ******************************************************************************
  */
/* USER CODE END Header */

/* Includes ------------------------------------------------------------------*/
#include "main.h"

/* Private includes ----------------------------------------------------------*/
/* USER CODE BEGIN Includes */
#include <string.h>
#include <stdlib.h>
#include <stdbool.h>
#include "si5351.h"
/* USER CODE END Includes */

/* Private typedef -----------------------------------------------------------*/
/* USER CODE BEGIN PTD */

/* USER CODE END PTD */

/* Private define ------------------------------------------------------------*/
/* USER CODE BEGIN PD */

/* USER CODE END PD */

/* Private macro -------------------------------------------------------------*/
/* USER CODE BEGIN PM */

/* USER CODE END PM */

/* Private variables ---------------------------------------------------------*/
ADC_HandleTypeDef hadc1;

I2C_HandleTypeDef hi2c1;

TIM_HandleTypeDef htim1;

/* USER CODE BEGIN PV */
/* Private variables ---------------------------------------------------------*/

/* USER CODE END PV */

/* Private function prototypes -----------------------------------------------*/
void SystemClock_Config(void);
static void MX_GPIO_Init(void);
static void MX_TIM1_Init(void);
static void MX_I2C1_Init(void);
static void MX_ADC1_Init(void);
/* USER CODE BEGIN PFP */
/* Private function prototypes -----------------------------------------------*/

/* USER CODE END PFP */

/* Private user code ---------------------------------------------------------*/
/* USER CODE BEGIN 0 */

#define CH_CW  0
#define CH_VFO 1
#define CH_BFO 2

// note: can be affected a little by the actual load in the circuit
const int32_t XtalFilterCenterFrequency = 8998433;
const int32_t si5351_correction = 5430;
int32_t Fvfo = 1000000; // dummy value, will be changed below

typedef enum {
	ROTARY_ENCODER_MODE_CHANGE_FREQUENCY = 0,
	ROTARY_ENCODER_MODE_CHANGE_KEYER_SPEED = 1,
} RotaryEncoderMode_t;

int32_t prevCounter = 0;
int32_t lastFrequency40M = 0;
int32_t lastFrequency20M = 14060000;
uint8_t currentBand = 40;
int32_t targetFrequency = 7030000;
bool prepareFastModeChange = false;
bool fastMode = false;
RotaryEncoderMode_t rotaryEncoderMode;

int32_t sendStart = 0;
int32_t sendFinish = 0;
int32_t keyReadTimeout = 0;
int32_t sendTimeout = 0;
int32_t keyerSpeedWPM = 18;
// keyerDitTimeMs = 60*1000/(50*WPM)
// where 50 is the length in dits of "PARIS "
// see https://morsecode.world/international/timing.html
int32_t keyerDitTimeMs = 66;
bool straightKeyer = false;
bool isSending = false;

typedef enum {
	CW_SEND_NONE = 0,
	CW_SEND_DIT = 1,
	CW_SEND_DAH = 2,
} CwSend_t;

CwSend_t lastSent = CW_SEND_NONE;
CwSend_t nextSend = CW_SEND_NONE;

#define PIN_RS    (1 << 0)
#define PIN_EN    (1 << 2)
#define BACKLIGHT (1 << 3)

#define LCD_DELAY_MS 5

uint8_t lcd_addr = (0x27 << 1);

#define BUTTON_ROT_ENCODER 0
#define BUTTON_DIT 1
#define BUTTON_DAH 2
bool inTransmitMode = false;
uint32_t transmitModeEnterTime = 0;

#define ROT_ENCODER_BUTTON_DEBOUNCE_TIME_MS 150
typedef enum {
	ROTARY_ENCODER_BUTTON_PRESSED = 0,
	ROTARY_ENCODER_BUTTON_RELEASED = 1,
	ROTARY_ENCODER_BUTTON_DEBOUNCE = 2,
} RotaryEncoderButtonStatus_t;

HAL_StatusTypeDef LCD_SendInternal(uint8_t data, uint8_t flags) {
    HAL_StatusTypeDef res;
    for(;;) {
        res = HAL_I2C_IsDeviceReady(&hi2c1, lcd_addr, 1,
                                    HAL_MAX_DELAY);
        if(res == HAL_OK)
            break;
    }

    uint8_t up = data & 0xF0;
    uint8_t lo = (data << 4) & 0xF0;

    uint8_t data_arr[4];
    data_arr[0] = up|flags|BACKLIGHT|PIN_EN;
    data_arr[1] = up|flags|BACKLIGHT;
    data_arr[2] = lo|flags|BACKLIGHT|PIN_EN;
    data_arr[3] = lo|flags|BACKLIGHT;

    res = HAL_I2C_Master_Transmit(&hi2c1, lcd_addr, data_arr,
                                  sizeof(data_arr), HAL_MAX_DELAY);
    HAL_Delay(LCD_DELAY_MS);
    return res;
}

void LCD_SendCommand(uint8_t cmd) {
    LCD_SendInternal(cmd, 0);
}

void LCD_SendData(uint8_t data) {
    LCD_SendInternal(data, PIN_RS);
}

void LCD_Init() {
    // 4-bit mode, 2 lines, 5x7 format
    LCD_SendCommand(0b00110000);
    // display & cursor home (keep this!)
    LCD_SendCommand(0b00000010);
    // display on, right shift, underline off, blink off
    LCD_SendCommand(0b00001100);
    // clear display (optional here)
    LCD_SendCommand(0b00000001);
}

void LCD_SendString(char *str) {
    while(*str) {
        LCD_SendData((uint8_t)(*str));
        str++;
    }
}

void SetupCLK(uint8_t output, int32_t Fclk, si5351DriveStrength_t driveStrength) {
	static bool pllSetupDone = false;
	si5351PLLConfig_t pll_conf;
	si5351OutputConfig_t out_conf;

	si5351_Calc(Fclk, &pll_conf, &out_conf);
	if(!pllSetupDone) {
		si5351_SetupPLL(SI5351_PLL_A, &pll_conf);
		pllSetupDone = true;
	}
	si5351_SetupOutput(output, SI5351_PLL_A, driveStrength, &out_conf);
}

void changeKeyerSpeed(int32_t delta) {
	keyerSpeedWPM += delta;
	if(keyerSpeedWPM < 10) {
		keyerSpeedWPM = 9;
		straightKeyer = true;
	} else {
		straightKeyer = false;
		if(keyerSpeedWPM > 35) {
			keyerSpeedWPM = 35;
		}
	}

	keyerDitTimeMs = 60*1000/(50*keyerSpeedWPM);
}

void changeFrequency(int32_t delta) {
	if(fastMode) {
		delta *= 10;
	}
	targetFrequency += 100*delta;
	if(fastMode && (delta != 0)) {
		targetFrequency -= targetFrequency % 1000;
	}

	if(currentBand == 40) {
		if(targetFrequency < 7000000) {
			targetFrequency = 7000000;
		} else if(targetFrequency > 7040000) {
			targetFrequency = 7040000;
		}
	} else { // 20m
		if(targetFrequency < 14000000) {
			targetFrequency = 14000000;
		} else if(targetFrequency > 14070000) {
			targetFrequency = 14070000;
		}
	}

	// use lowest VFO possible
	if(targetFrequency > XtalFilterCenterFrequency) {
		Fvfo = targetFrequency - XtalFilterCenterFrequency;
	} else {
		Fvfo = XtalFilterCenterFrequency - targetFrequency;
	}

	si5351_EnableOutputs((1 << CH_BFO));
	SetupCLK(CH_VFO, Fvfo, SI5351_DRIVE_STRENGTH_2MA);
	si5351_EnableOutputs((1 << CH_VFO)|(1 << CH_BFO));
}

void changeRotaryEncoderMode() {
	if(rotaryEncoderMode == ROTARY_ENCODER_MODE_CHANGE_FREQUENCY) {
		rotaryEncoderMode = ROTARY_ENCODER_MODE_CHANGE_KEYER_SPEED;
	} else {
		rotaryEncoderMode = ROTARY_ENCODER_MODE_CHANGE_FREQUENCY;
	}
}

void displayKeyerSettings() {
	char buff[16];
	if(straightKeyer) {
		LCD_SendCommand(0b10000000);
		LCD_SendString("SPEED --");
		LCD_SendCommand(0b11000000);
		LCD_SendString("STRAIGHT");
	} else {
		snprintf(buff, sizeof(buff), "SPEED %02ld", keyerSpeedWPM);
		LCD_SendCommand(0b10000000);
		LCD_SendString(buff);
		LCD_SendCommand(0b11000000);
		LCD_SendString("IAMBIC  ");
	}
}

void displayFrequency() {
	char buff[16];
	snprintf(buff, sizeof(buff), "%02ld.%03ld.%01ld", 
		targetFrequency / 1000000, (targetFrequency % 1000000) / 1000, (targetFrequency % 1000) / 100);

    LCD_SendCommand(0b10000000);
    LCD_SendString(buff);
}

void keyDown() {
	si5351_EnableOutputs((1 << CH_VFO)|(1 << CH_BFO)|(1 << CH_CW));
}

void keyUp() {
	si5351_EnableOutputs((1 << CH_VFO)|(1 << CH_BFO));
}

void switchRelayTx() {
	HAL_GPIO_WritePin(GPIOA, GPIO_PIN_0, GPIO_PIN_SET);
	HAL_Delay(10);
	HAL_GPIO_WritePin(GPIOA, GPIO_PIN_0, GPIO_PIN_RESET);
}

void switchRelayRx() {
	HAL_GPIO_WritePin(GPIOC, GPIO_PIN_15, GPIO_PIN_SET);
	HAL_Delay(10);
	HAL_GPIO_WritePin(GPIOC, GPIO_PIN_15, GPIO_PIN_RESET);
}

void switchRelay20M() {
	HAL_GPIO_WritePin(GPIOC, GPIO_PIN_14, GPIO_PIN_SET);
	HAL_Delay(10);
	HAL_GPIO_WritePin(GPIOC, GPIO_PIN_14, GPIO_PIN_RESET);
}

void switchRelay40M() {
	HAL_GPIO_WritePin(GPIOC, GPIO_PIN_13, GPIO_PIN_SET);
	HAL_Delay(10);
	HAL_GPIO_WritePin(GPIOC, GPIO_PIN_13, GPIO_PIN_RESET);
}

void changeBand() {
	if(currentBand == 40) {
		lastFrequency40M = targetFrequency;
		targetFrequency = lastFrequency20M;
		currentBand = 20;
		switchRelay20M();
	} else { // 20m
		lastFrequency20M = targetFrequency;
		targetFrequency = lastFrequency40M;
		currentBand = 40;
		switchRelay40M();
	}
	changeFrequency(0);
	displayFrequency();
}

void ensureTransmitMode() {
	if(inTransmitMode) {
		return;
	}

	si5351_EnableOutputs(0);
	// attenuate our own received signal by shifting VFO/BFO frequencies
	int32_t delta = 10000;
	if(targetFrequency > XtalFilterCenterFrequency) { 
		SetupCLK(CH_VFO, Fvfo-delta, SI5351_DRIVE_STRENGTH_2MA);
	} else {
		SetupCLK(CH_VFO, Fvfo+delta, SI5351_DRIVE_STRENGTH_2MA);
	}
	SetupCLK(CH_BFO, XtalFilterCenterFrequency + 1000 + delta, SI5351_DRIVE_STRENGTH_2MA);
	SetupCLK(CH_CW, targetFrequency, SI5351_DRIVE_STRENGTH_2MA);
	si5351_EnableOutputs((1 << CH_VFO)|(1 << CH_BFO));

	// Power ON the PA driver.
	// Experiments showed that doing it here instead of keyUp()/keyDown()
	// produces better sounding signal. Also the signal sounds better
	// comparing to keeping CH_CW always ON and keying it with the PA driver.
	HAL_GPIO_WritePin(GPIOB, GPIO_PIN_0, GPIO_PIN_SET);

	// Some oscillations are possible at this moment that can cause extra clicks,
	// so keep RX state a little longer
	HAL_Delay(10);
	LCD_SendCommand(0b11000000);
	LCD_SendString("TX");
	switchRelayTx();

	inTransmitMode = true;
}

void ensureReceiveMode() {
	if(!inTransmitMode) {
		return;
	}

	// Power OFF the PA driver.
	HAL_GPIO_WritePin(GPIOB, GPIO_PIN_0, GPIO_PIN_RESET);

	// CH_CW can be heard in the headphones which is a bit annoying
	// thus CH_CW is keyed in keyDown()/keyUp()
	// si5351_EnableOutputs((1 << CH_VFO)|(1 << CH_BFO));

	// Restore the original VFO and BFO
	si5351_EnableOutputs(0);
	SetupCLK(CH_VFO, Fvfo, SI5351_DRIVE_STRENGTH_2MA);
	SetupCLK(CH_BFO, XtalFilterCenterFrequency + 1000, SI5351_DRIVE_STRENGTH_2MA);
	si5351_EnableOutputs((1 << CH_VFO)|(1 << CH_BFO));

	// Some oscillations are possible at this moment that can trigger AGC,
	// so keep TX state a little longer
	HAL_Delay(10);
	LCD_SendCommand(0b11000000);
	LCD_SendString("RX");
	switchRelayRx();

	inTransmitMode = false;
}

void init() {
/*
	// Code for determining the correction factor for Si5351

	si5351_Init(0);
	si5351PLLConfig_t pll_conf;
	si5351OutputConfig_t out_conf;
	int32_t Fclk = 10000000;

	si5351_Calc(Fclk, &pll_conf, &out_conf);
	si5351_SetupPLL(SI5351_PLL_A, &pll_conf);
	si5351_SetupOutput(0, SI5351_PLL_A, SI5351_DRIVE_STRENGTH_2MA, &out_conf);
	si5351_EnableOutputs(1<<0);

	while(1) {
		HAL_Delay(100);
	}
*/

	keyUp();
	switchRelayRx();
	switchRelay40M();

	// Give the LCD and Si5351 some time to initialize after powering up
	HAL_Delay(1000);

	// Check for an alternative LCD address
    HAL_StatusTypeDef res;
    res = HAL_I2C_IsDeviceReady(&hi2c1, (0x3F << 1), 1, HAL_MAX_DELAY);
    if(res == HAL_OK) {
		lcd_addr = (0x3F << 1);
	}

	HAL_TIM_Encoder_Start(&htim1, TIM_CHANNEL_ALL);
	HAL_ADC_Start(&hadc1);

    LCD_Init();
    LCD_SendCommand(0b10000000);
    LCD_SendString(" AYN /P ");
    LCD_SendCommand(0b11000000);
    LCD_SendString("Oct 2021");
	HAL_Delay(1000);
    // clear display
    LCD_SendCommand(0b00000001);

    LCD_SendCommand(0b10000000);
    LCD_SendString("I2C Scan:");
    LCD_SendCommand(0b11000000);
    for(uint16_t i = 0; i < 128; i++) {
		HAL_StatusTypeDef res;
        res = HAL_I2C_IsDeviceReady(&hi2c1, i << 1, 1, 10);
        if(res == HAL_OK) {
            char msg[64];
            snprintf(msg, sizeof(msg), "%02X ", i);
    		LCD_SendString(msg);
        }
    }  
	HAL_Delay(1000);
    // clear display
    LCD_SendCommand(0b00000001);

	LCD_SendCommand(0b11000000);
	LCD_SendString("RX");

	si5351_Init(si5351_correction);
	SetupCLK(CH_BFO, XtalFilterCenterFrequency + 1000, SI5351_DRIVE_STRENGTH_2MA);
	changeFrequency(0);
  
    // display on, right shift, underline off, blink off
    LCD_SendCommand(0b00001100);
	displayFrequency();
}

bool buttonRotEncoderLongPressed() {
	static uint32_t pressedSince = 0;
	uint32_t now = HAL_GetTick();
	if(HAL_GPIO_ReadPin(GPIOB, GPIO_PIN_10) == GPIO_PIN_RESET) {
		if(pressedSince == 0) {
			pressedSince = now;
		}
	} else {
		pressedSince = 0;
	}

	return (pressedSince > 0) && (now - pressedSince > 450);
}

RotaryEncoderButtonStatus_t buttonRotEncoderPressed() {
	static uint32_t lastPressed = 0;
	if(HAL_GPIO_ReadPin(GPIOB, GPIO_PIN_10) == GPIO_PIN_RESET) {
		uint32_t now = HAL_GetTick();
		if(now - lastPressed > ROT_ENCODER_BUTTON_DEBOUNCE_TIME_MS) {
			lastPressed = now;
			return ROTARY_ENCODER_BUTTON_PRESSED;
		} else {
			return ROTARY_ENCODER_BUTTON_DEBOUNCE;
		}
	}
	return ROTARY_ENCODER_BUTTON_RELEASED;
}

bool buttonDitPressed() {
	return (HAL_GPIO_ReadPin(GPIOB, GPIO_PIN_11) == GPIO_PIN_RESET);
}

bool buttonDahPressed() {
	return (HAL_GPIO_ReadPin(GPIOB, GPIO_PIN_12) == GPIO_PIN_RESET);
}

void processStraightKeyerLogic(bool pressed) {
	if(pressed) {
		keyDown();
	} else {
		keyUp();
	}
}

void processIambicKeyerLogic(bool ditPressed, bool dahPressed) {
	int32_t now = HAL_GetTick();

	if(isSending && (now >= sendFinish)) {
		keyUp();
		isSending = false;
	}

	if(nextSend == CW_SEND_NONE) {
		if(ditPressed && (lastSent == CW_SEND_DAH)) {
			nextSend = CW_SEND_DIT;
		} else if(dahPressed && (lastSent == CW_SEND_DIT)) {
			nextSend = CW_SEND_DAH;
		} else if(now > keyReadTimeout) {
			if(ditPressed) {
				nextSend = CW_SEND_DIT;
			} else if (dahPressed) {
				nextSend = CW_SEND_DAH;
			}
		}
	}

	if((now > sendTimeout) && (nextSend != CW_SEND_NONE)) {
		if(nextSend == CW_SEND_DIT) {
			sendFinish = now + keyerDitTimeMs;
			keyReadTimeout = now + keyerDitTimeMs*2;
		} else if(nextSend == CW_SEND_DAH) {
			sendFinish = now + keyerDitTimeMs*3;
			keyReadTimeout = now + keyerDitTimeMs*3;
		}

		lastSent = nextSend;
		nextSend = CW_SEND_NONE;
		sendStart = now;
		sendTimeout = sendFinish + keyerDitTimeMs;
		isSending = true;
		keyDown();
	}
}

void displayVoltageOrMode(int32_t adcValue, bool force) {
	// don't forget to add '-u _printf_float' to LDFLAGS
	char buff[16];
    static double prevVoltage = 0.0;
	double voltage = ((double)(adcValue)-1220.0)/190.0 + 6.9;

	if(force && fastMode) {
		LCD_SendCommand(0b11000000);
		LCD_SendString("RX FAST ");
		return;
	}

    if((!force) && (abs((int)(voltage*10) - (int)(prevVoltage*10)) <= 1)) {
		return;
	}

	prevVoltage = voltage;

	if(voltage < 9.95) {
		snprintf(buff, sizeof(buff), "RX  %.1fV", voltage);
	} else {
		snprintf(buff, sizeof(buff), "RX %.1fV", voltage);
	}

	LCD_SendCommand(0b11000000);
	LCD_SendString(buff);
}

void loop() {
	static uint32_t lastStateChange = 0;
	static int32_t lastVoltageUpdate = 0;
	static int32_t lastVoltageValue = 1234;
	bool ditPressed = buttonDitPressed();
	bool dahPressed = buttonDahPressed();
	if((rotaryEncoderMode == ROTARY_ENCODER_MODE_CHANGE_FREQUENCY) && (ditPressed || dahPressed)) {
		transmitModeEnterTime = HAL_GetTick();
		ensureTransmitMode();
	} else {
		uint32_t tstamp = HAL_GetTick();
		if(tstamp - transmitModeEnterTime > keyerDitTimeMs*10) {
			ensureReceiveMode();
		}
	}

	int32_t currCounter = __HAL_TIM_GET_COUNTER(&htim1);
	currCounter = 32767 - ((currCounter-1) & 0xFFFF) / 2;
	if(currCounter > 32768/2) {
		// convert ... 32766, 32767, 0, 1, 2 ... into:
		//               ... -2, -1, 0, 1, 2 ...
		// this simplifies `delta` calculation
		currCounter = currCounter-32768;
	}
	if(currCounter != prevCounter) {
		int32_t delta = currCounter-prevCounter;
		prevCounter = currCounter;
		// debounce or skip counter overflow
		if((delta > -10) && (delta < 10) && (!inTransmitMode)) {
			delta = -delta; // the encoder was soldered backwards ¯\_(ツ)_/¯
			if(rotaryEncoderMode == ROTARY_ENCODER_MODE_CHANGE_FREQUENCY) {
				changeFrequency(delta);
				displayFrequency();
			} else {
				changeKeyerSpeed(delta);
				displayKeyerSettings();
			}
		}
	}

	if(inTransmitMode) {
		if(straightKeyer) {
			processStraightKeyerLogic(ditPressed);
		} else {
			processIambicKeyerLogic(ditPressed, dahPressed);
		}
	} else {
		uint32_t now = HAL_GetTick();
		if(prepareFastModeChange && (now - lastStateChange > 500)) {
			// rotary encoder was pressed, and it was not a double click, nor a long press
			prepareFastModeChange = false;
			fastMode = !fastMode;
			displayVoltageOrMode(lastVoltageValue, true);
		}

		if((!fastMode) && (now - lastVoltageUpdate > 1000)) {
		    HAL_ADC_PollForConversion(&hadc1, HAL_MAX_DELAY);
			lastVoltageValue = HAL_ADC_GetValue(&hadc1);
			displayVoltageOrMode(lastVoltageValue, false);
			lastVoltageUpdate = now;
		}

		if(buttonRotEncoderLongPressed()) {
			if(rotaryEncoderMode == ROTARY_ENCODER_MODE_CHANGE_FREQUENCY) {
				prepareFastModeChange = false;
				changeRotaryEncoderMode();
				displayKeyerSettings();
			} // do nothing in keyer menu
		} else {
			static uint8_t changeStateTimes = 0;
			static RotaryEncoderButtonStatus_t prevStatus = ROTARY_ENCODER_BUTTON_RELEASED;
			RotaryEncoderButtonStatus_t status = buttonRotEncoderPressed();

			if(status == ROTARY_ENCODER_BUTTON_PRESSED) {
				if(rotaryEncoderMode == ROTARY_ENCODER_MODE_CHANGE_FREQUENCY) {
					// first make sure it wasn't a long press or a double click
					prepareFastModeChange = true;
				} else {
					// leave keyer menu
					changeRotaryEncoderMode();
					displayFrequency();
					displayVoltageOrMode(lastVoltageValue, true);
				}
			}

			if((status != prevStatus) && (status != ROTARY_ENCODER_BUTTON_DEBOUNCE)) {
				if(now - lastStateChange < 500) { 
					changeStateTimes++;
				} else {
					changeStateTimes = 0;
				}
				lastStateChange = now;
				prevStatus = status;
			}

			if(changeStateTimes >= 3) { // double click
				prepareFastModeChange = false;
				changeBand();
				rotaryEncoderMode = ROTARY_ENCODER_MODE_CHANGE_FREQUENCY;
				changeStateTimes = 0;
				lastStateChange = 0;
				prevStatus = ROTARY_ENCODER_BUTTON_RELEASED;
			}
		}
	}

    HAL_Delay(5);
}

/* USER CODE END 0 */

/**
  * @brief  The application entry point.
  * @retval int
  */
int main(void)
{
  /* USER CODE BEGIN 1 */

  /* USER CODE END 1 */

  /* MCU Configuration--------------------------------------------------------*/

  /* Reset of all peripherals, Initializes the Flash interface and the Systick. */
  HAL_Init();

  /* USER CODE BEGIN Init */

  /* USER CODE END Init */

  /* Configure the system clock */
  SystemClock_Config();

  /* USER CODE BEGIN SysInit */

  /* USER CODE END SysInit */

  /* Initialize all configured peripherals */
  MX_GPIO_Init();
  MX_TIM1_Init();
  MX_I2C1_Init();
  MX_ADC1_Init();
  /* USER CODE BEGIN 2 */
  init();
  /* USER CODE END 2 */

  /* Infinite loop */
  /* USER CODE BEGIN WHILE */
  while (1)
  {
    loop();
    /* USER CODE END WHILE */

    /* USER CODE BEGIN 3 */

  }
  /* USER CODE END 3 */
}

/**
  * @brief System Clock Configuration
  * @retval None
  */
void SystemClock_Config(void)
{
  RCC_OscInitTypeDef RCC_OscInitStruct = {0};
  RCC_ClkInitTypeDef RCC_ClkInitStruct = {0};
  RCC_PeriphCLKInitTypeDef PeriphClkInit = {0};

  /**Initializes the CPU, AHB and APB busses clocks 
  */
  RCC_OscInitStruct.OscillatorType = RCC_OSCILLATORTYPE_HSI;
  RCC_OscInitStruct.HSIState = RCC_HSI_ON;
  RCC_OscInitStruct.HSICalibrationValue = RCC_HSICALIBRATION_DEFAULT;
  RCC_OscInitStruct.PLL.PLLState = RCC_PLL_NONE;
  if (HAL_RCC_OscConfig(&RCC_OscInitStruct) != HAL_OK)
  {
    Error_Handler();
  }
  /**Initializes the CPU, AHB and APB busses clocks 
  */
  RCC_ClkInitStruct.ClockType = RCC_CLOCKTYPE_HCLK|RCC_CLOCKTYPE_SYSCLK
                              |RCC_CLOCKTYPE_PCLK1|RCC_CLOCKTYPE_PCLK2;
  RCC_ClkInitStruct.SYSCLKSource = RCC_SYSCLKSOURCE_HSI;
  RCC_ClkInitStruct.AHBCLKDivider = RCC_SYSCLK_DIV1;
  RCC_ClkInitStruct.APB1CLKDivider = RCC_HCLK_DIV1;
  RCC_ClkInitStruct.APB2CLKDivider = RCC_HCLK_DIV1;

  if (HAL_RCC_ClockConfig(&RCC_ClkInitStruct, FLASH_LATENCY_0) != HAL_OK)
  {
    Error_Handler();
  }
  PeriphClkInit.PeriphClockSelection = RCC_PERIPHCLK_ADC;
  PeriphClkInit.AdcClockSelection = RCC_ADCPCLK2_DIV2;
  if (HAL_RCCEx_PeriphCLKConfig(&PeriphClkInit) != HAL_OK)
  {
    Error_Handler();
  }
}

/**
  * @brief ADC1 Initialization Function
  * @param None
  * @retval None
  */
static void MX_ADC1_Init(void)
{

  /* USER CODE BEGIN ADC1_Init 0 */

  /* USER CODE END ADC1_Init 0 */

  ADC_ChannelConfTypeDef sConfig = {0};

  /* USER CODE BEGIN ADC1_Init 1 */

  /* USER CODE END ADC1_Init 1 */
  /**Common config 
  */
  hadc1.Instance = ADC1;
  hadc1.Init.ScanConvMode = ADC_SCAN_DISABLE;
  hadc1.Init.ContinuousConvMode = ENABLE;
  hadc1.Init.DiscontinuousConvMode = DISABLE;
  hadc1.Init.ExternalTrigConv = ADC_SOFTWARE_START;
  hadc1.Init.DataAlign = ADC_DATAALIGN_RIGHT;
  hadc1.Init.NbrOfConversion = 1;
  if (HAL_ADC_Init(&hadc1) != HAL_OK)
  {
    Error_Handler();
  }
  /**Configure Regular Channel 
  */
  sConfig.Channel = ADC_CHANNEL_1;
  sConfig.Rank = ADC_REGULAR_RANK_1;
  sConfig.SamplingTime = ADC_SAMPLETIME_1CYCLE_5;
  if (HAL_ADC_ConfigChannel(&hadc1, &sConfig) != HAL_OK)
  {
    Error_Handler();
  }
  /* USER CODE BEGIN ADC1_Init 2 */

  /* USER CODE END ADC1_Init 2 */

}

/**
  * @brief I2C1 Initialization Function
  * @param None
  * @retval None
  */
static void MX_I2C1_Init(void)
{

  /* USER CODE BEGIN I2C1_Init 0 */

  /* USER CODE END I2C1_Init 0 */

  /* USER CODE BEGIN I2C1_Init 1 */

  /* USER CODE END I2C1_Init 1 */
  hi2c1.Instance = I2C1;
  hi2c1.Init.ClockSpeed = 100000;
  hi2c1.Init.DutyCycle = I2C_DUTYCYCLE_2;
  hi2c1.Init.OwnAddress1 = 0;
  hi2c1.Init.AddressingMode = I2C_ADDRESSINGMODE_7BIT;
  hi2c1.Init.DualAddressMode = I2C_DUALADDRESS_DISABLE;
  hi2c1.Init.OwnAddress2 = 0;
  hi2c1.Init.GeneralCallMode = I2C_GENERALCALL_DISABLE;
  hi2c1.Init.NoStretchMode = I2C_NOSTRETCH_DISABLE;
  if (HAL_I2C_Init(&hi2c1) != HAL_OK)
  {
    Error_Handler();
  }
  /* USER CODE BEGIN I2C1_Init 2 */

  /* USER CODE END I2C1_Init 2 */

}

/**
  * @brief TIM1 Initialization Function
  * @param None
  * @retval None
  */
static void MX_TIM1_Init(void)
{

  /* USER CODE BEGIN TIM1_Init 0 */

  /* USER CODE END TIM1_Init 0 */

  TIM_Encoder_InitTypeDef sConfig = {0};
  TIM_MasterConfigTypeDef sMasterConfig = {0};

  /* USER CODE BEGIN TIM1_Init 1 */

  /* USER CODE END TIM1_Init 1 */
  htim1.Instance = TIM1;
  htim1.Init.Prescaler = 0;
  htim1.Init.CounterMode = TIM_COUNTERMODE_UP;
  htim1.Init.Period = 65535;
  htim1.Init.ClockDivision = TIM_CLOCKDIVISION_DIV1;
  htim1.Init.RepetitionCounter = 0;
  htim1.Init.AutoReloadPreload = TIM_AUTORELOAD_PRELOAD_DISABLE;
  sConfig.EncoderMode = TIM_ENCODERMODE_TI1;
  sConfig.IC1Polarity = TIM_ICPOLARITY_RISING;
  sConfig.IC1Selection = TIM_ICSELECTION_DIRECTTI;
  sConfig.IC1Prescaler = TIM_ICPSC_DIV1;
  sConfig.IC1Filter = 0;
  sConfig.IC2Polarity = TIM_ICPOLARITY_RISING;
  sConfig.IC2Selection = TIM_ICSELECTION_DIRECTTI;
  sConfig.IC2Prescaler = TIM_ICPSC_DIV1;
  sConfig.IC2Filter = 0;
  if (HAL_TIM_Encoder_Init(&htim1, &sConfig) != HAL_OK)
  {
    Error_Handler();
  }
  sMasterConfig.MasterOutputTrigger = TIM_TRGO_RESET;
  sMasterConfig.MasterSlaveMode = TIM_MASTERSLAVEMODE_DISABLE;
  if (HAL_TIMEx_MasterConfigSynchronization(&htim1, &sMasterConfig) != HAL_OK)
  {
    Error_Handler();
  }
  /* USER CODE BEGIN TIM1_Init 2 */

  /* USER CODE END TIM1_Init 2 */

}

/**
  * @brief GPIO Initialization Function
  * @param None
  * @retval None
  */
static void MX_GPIO_Init(void)
{
  GPIO_InitTypeDef GPIO_InitStruct = {0};

  /* GPIO Ports Clock Enable */
  __HAL_RCC_GPIOC_CLK_ENABLE();
  __HAL_RCC_GPIOA_CLK_ENABLE();
  __HAL_RCC_GPIOB_CLK_ENABLE();

  /*Configure GPIO pin Output Level */
  HAL_GPIO_WritePin(GPIOC, GPIO_PIN_13|GPIO_PIN_14|GPIO_PIN_15, GPIO_PIN_RESET);

  /*Configure GPIO pin Output Level */
  HAL_GPIO_WritePin(GPIOA, GPIO_PIN_0, GPIO_PIN_RESET);

  /*Configure GPIO pin Output Level */
  HAL_GPIO_WritePin(GPIOB, GPIO_PIN_0, GPIO_PIN_RESET);

  /*Configure GPIO pins : PC13 PC14 PC15 */
  GPIO_InitStruct.Pin = GPIO_PIN_13|GPIO_PIN_14|GPIO_PIN_15;
  GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_PP;
  GPIO_InitStruct.Pull = GPIO_NOPULL;
  GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_LOW;
  HAL_GPIO_Init(GPIOC, &GPIO_InitStruct);

  /*Configure GPIO pin : PA0 */
  GPIO_InitStruct.Pin = GPIO_PIN_0;
  GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_PP;
  GPIO_InitStruct.Pull = GPIO_NOPULL;
  GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_LOW;
  HAL_GPIO_Init(GPIOA, &GPIO_InitStruct);

  /*Configure GPIO pin : PB0 */
  GPIO_InitStruct.Pin = GPIO_PIN_0;
  GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_PP;
  GPIO_InitStruct.Pull = GPIO_NOPULL;
  GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_LOW;
  HAL_GPIO_Init(GPIOB, &GPIO_InitStruct);

  /*Configure GPIO pins : PB10 PB11 PB12 */
  GPIO_InitStruct.Pin = GPIO_PIN_10|GPIO_PIN_11|GPIO_PIN_12;
  GPIO_InitStruct.Mode = GPIO_MODE_IT_RISING_FALLING;
  GPIO_InitStruct.Pull = GPIO_PULLUP;
  HAL_GPIO_Init(GPIOB, &GPIO_InitStruct);

  /* EXTI interrupt init*/
  HAL_NVIC_SetPriority(EXTI15_10_IRQn, 0, 0);
  HAL_NVIC_EnableIRQ(EXTI15_10_IRQn);

}

/* USER CODE BEGIN 4 */

/* USER CODE END 4 */

/**
  * @brief  This function is executed in case of error occurrence.
  * @retval None
  */
void Error_Handler(void)
{
  /* USER CODE BEGIN Error_Handler_Debug */
  /* User can add his own implementation to report the HAL error return state */
  while(1) 
  {
  }
  /* USER CODE END Error_Handler_Debug */
}

#ifdef  USE_FULL_ASSERT
/**
  * @brief  Reports the name of the source file and the source line number
  *         where the assert_param error has occurred.
  * @param  file: pointer to the source file name
  * @param  line: assert_param error line source number
  * @retval None
  */
void assert_failed(uint8_t *file, uint32_t line)
{ 
  /* USER CODE BEGIN 6 */
  /* USER CODE END 6 */
}
#endif /* USE_FULL_ASSERT */

/************************ (C) COPYRIGHT STMicroelectronics *****END OF FILE****/
